//
// verification routines (for new specification style).
// 
//

// Three major functions:   
// 1) setting up and managing the table of regions
// 2) controlling the execution of statements
// 3) evaluating expressions
//  


#include "exec_spec.H"

// ==================== setting up table ==============================

// initialize the reg table reg_table, given reg_name_table.
// no need to store the names, just NULL pointers, and set size
void 
set_up_reg_table(region_table *reg_table, 
		      struct region_name_table *reg_name_table) {
  int i;
  reg_table->size = reg_name_table->size;
  for (i=0;i<reg_table->size;i++) {
    reg_table->regions[i] = (region_array *) NULL;
  }
}



// ==================== evaluating expressions ========================

// returns a fresh region_array for the expression
// any intermediate expressions used are freed
region_array *
evaluate_expr(struct expression *expr, 
	      struct region_table *reg_table, 
	      automaton *aut) {
//  partially implemented
  region_array *rap1, *rap2, *oldrap;
  region_array *result_rap, *init_r;
  Bool converged;
  int iter_count = 0;

  // for each expr_type, set the array result_rap 
  switch (expr->type) {
  case REGION_NAME:
    result_rap = new region_array(reg_table->regions[expr->rindex]);
    break;
  case AND: 
    rap1 = evaluate_expr(expr->expr1,reg_table,aut);
    rap2 = evaluate_expr(expr->expr2,reg_table,aut);
    result_rap = rap1->ra_and(*rap2);
// result_rap->print(); cout << "---\n" << flush;
    rap1->free(); global_free(rap1);
    rap2->free(); global_free(rap2);
    break;
  case OR: 
    rap1 = evaluate_expr(expr->expr1,reg_table,aut);
    rap2 = evaluate_expr(expr->expr2,reg_table,aut);
    // destroys rap1
    result_rap = rap1->ra_or(*rap2);
    rap2->free(); global_free(rap2);
    break;
  case WEAKDIFF: 
    rap1 = evaluate_expr(expr->expr1,reg_table,aut);
    rap2 = evaluate_expr(expr->expr2,reg_table,aut);
    // destroys rap1
    // obtain the result as by product of the "new" part of ra_or_new
    rap2->ra_or_new(*rap1,&result_rap);
    rap1->free(); global_free(rap1);
    rap2->free(); global_free(rap2);
    break;
  case DIFF: 
    rap1 = evaluate_expr(expr->expr1,reg_table,aut);
    rap2 = evaluate_expr(expr->expr2,reg_table,aut);
    result_rap = rap1->ra_minus(rap2);
    rap1->free(); global_free(rap1);
    rap2->free(); global_free(rap2);
    break;
  case COMPLEMENT:
    rap1 = evaluate_expr(expr->expr1,reg_table,aut);
    result_rap = rap1->ra_complement();
    rap1->free(); global_free(rap1);
    break;
  case PRE_EXPR:
    rap1  = evaluate_expr(expr->expr1,reg_table,aut);
    if (!aut->pre_is_valid) {
      aut->set_pre();
    }
    rap1->restrict_to_invariant();
    result_rap = rap1->ra_full_pre();
    result_rap->simplify();
    rap1->free(); global_free(rap1);
    break;
  case POST_EXPR:
    rap1  = evaluate_expr(expr->expr1,reg_table,aut);
    rap1->restrict_to_invariant();
    result_rap = rap1->ra_full_post();
    result_rap->simplify();
    rap1->free(); global_free(rap1);
    break;
  case PREDICATE:
    result_rap = new region_array(expr->pred, aut);
    break;
  case REACHFORWARD:
    rap1 = evaluate_expr(expr->expr1,reg_table,aut);
//    rap1->print_size();
    // currently using ra_reach with non-full pre/post
    init_r = new region_array(rap1);
    rap1->prepare_for_reach(POST,1);
    result_rap = rap1->ra_reach(POST);
    result_rap->init_regn = init_r;
    rap1->free(); global_free(rap1);
    break;
  case  REACHBACKWARD:
    if (!aut->pre_is_valid) {
      aut->set_pre();
    }
    rap1 = evaluate_expr(expr->expr1,reg_table,aut);
    init_r = new region_array(rap1);
    rap1->prepare_for_reach(PRE,1);
    result_rap = rap1->ra_reach(PRE);
    result_rap->init_regn = init_r;
    rap1->free(); global_free(rap1);
    break;
  case  RECURSION:
    // iterate on expr->rindex from expr->expr1, using expr->stmts
    // assign oldrap, then check at each iteration for weak  equivalence
    converged = false;
    oldrap = evaluate_expr(expr->expr1,reg_table,aut);

    reg_table->regions[expr->rindex] = new region_array(oldrap);
    while (!converged) {
      execute_statements(expr->stmt,reg_table, aut);
      converged = reg_table->regions[expr->rindex]->ra_equiv(oldrap);
      if (!converged) {
	oldrap->free(); global_free(oldrap);
	oldrap = new region_array(reg_table->regions[expr->rindex]);
      }
      cerr << "." << flush;
      iter_count++;
    }
    cerr << "Completed recursion in " << iter_count << " iterations\n" 
      << flush;
    cout << "Completed recursion in " << iter_count << " iterations\n" 
      << flush;
    result_rap = oldrap; // i.e. return fresh copy, 
                    // not necessarily reg_table->regions[expr->rindex];
    break;
  case HIDE_EXPR:
    rap1 = evaluate_expr(expr->expr1,reg_table,aut);
    switch (expr->elim->elim_all) {
    case 2:
      result_rap = rap1->eliminate_all_vars();
      break;
    case 1:
      result_rap = rap1->eliminate_non_params();
      break;
    case 0:
      result_rap = rap1->eliminate_vars(expr->elim->name_list);
      break;
    default:
      cerr << "Illegal call of hide\n" << flush; exit(1);
      break;
    }
   break;

 case OMIT_EXPR:
    rap1 = evaluate_expr(expr->expr1,reg_table,aut);
    elim_all_locns = 1;
    if (!elim_all_locns) {
      cout << "Cannot process omit statement unless eliminating all locns\n";
      cout << "Will eliminate ALL locns\n";
      elim_all_locns = 1;
    }
    elim_locs_by_list = ! elim_all_locns;
    elim_proc_list = NULL;
    // also sets the elim var info, but this is irrelevant and does no harm
    aut->set_elim_info();
    aut->copy_location_names();
    rap1->elim_locs_and_store(aut);
    result_rap = rap1;
    aut->restore_location_names();
    break;

  case HULL_EXPR:
    result_rap  = evaluate_expr(expr->expr1,reg_table,aut);
    result_rap = result_rap->ra_hull();
    break;
  default:
    cerr << "Aborting: Illegal expr type in call of eval expr\n" << flush;
    exit(1);
  } 
  return(result_rap);
}


// ==================== evaluating Boolean expressions ========================



Bool
result_of_Boolexpression(struct Boolexpression *Boolexpr, 
			 struct region_table *reg_table, 
			 automaton *aut) {
  Bool result;
  region_array *rap;
  switch (Boolexpr->type) {
  case BE_AND:
    result =    result_of_Boolexpression(Boolexpr->Boolexpr1,reg_table,aut)
             && result_of_Boolexpression(Boolexpr->Boolexpr2,reg_table,aut);
    break;
  case BE_OR:
    result =    result_of_Boolexpression(Boolexpr->Boolexpr1,reg_table,aut)
             || result_of_Boolexpression(Boolexpr->Boolexpr2,reg_table,aut);
    break;
  case BE_NOT:
    result =   ! result_of_Boolexpression(Boolexpr->Boolexpr1,reg_table,aut);
    break;
  case BE_EMPTY:
    rap =  evaluate_expr(Boolexpr->left_expr, reg_table, aut);
    result = rap->ra_empty();
    rap->free(); global_free(rap);
    break;
  case BE_WEAKLE : case BE_WEAKEQ: case BE_WEAKGE:
    {
      region_array *raleft = evaluate_expr(Boolexpr->left_expr,
					       reg_table,
					       aut);
      region_array *raright = evaluate_expr(Boolexpr->right_expr,
					       reg_table,
					       aut);
      switch (Boolexpr->type) {
      case BE_WEAKLE: 
	result = raleft->ra_in(raright);
	break;
      case BE_WEAKEQ: 
	result = raleft->ra_equiv(raright);
	break;
      case BE_WEAKGE: 
	result = raright->ra_in(raleft);
	break;
      default: break;
      } // end switch
    } // end of block for WEAK ops
    break;
  case BE_INEQ:
    {
    region_array *raleft = evaluate_expr(Boolexpr->left_expr,
					       reg_table,
					       aut);
    region_array *raright = evaluate_expr(Boolexpr->right_expr,
					       reg_table,
					       aut);
    switch(Boolexpr->optype) {
    case LE:
      result = raleft->ra_exact_is_in(raright);
      break;
    case EQ:
      result = raleft->ra_exact_is_equal(raright);
      break;
    case GE:
      result = raright->ra_exact_is_in(raleft);
      break;
    case NEQ:
      result = !raright->ra_exact_is_equal(raleft);
      break;
    case LT:
      result = raleft->ra_exact_is_lt(raright);
      break;
    case GT:
      result = raleft->ra_exact_is_gt(raright);
      break;
    default:
      cerr << "Illegal call to result_of_Boolexpression, with type " <<
	Boolexpr->optype << "\n";
      exit(1);
      break;
    } // end of switch of ineq type
    raleft->free(); global_free(raleft);
    raright->free(); global_free(raright);
    } 
    break;
  default:
    cerr << "Illegal Boolexpr->type " << Boolexpr->type << " in call of " <<
      "result_of_Boolexpression()\n" << flush;
    exit(1);
  } // end switch of Boolexpr type
  return(result);
}


// ==================== executing statements ==========================
// 
// blindly execute statements requested.
// beware. a region may be freed, and then a print attempted.
// 

// assumes that stmt is non-NULL
void
execute_statement(struct statement *stmt, 
			struct region_table *reg_table, automaton *aut) {
  region_array *rap;
  region_array *rap1, *rap2;

//   region * resultreg; // for simplified output

  switch (stmt->type) {
  case PRINT_STMT: // print the region_array
    /* // if a name only
    if (reg_table->regions[stmt->rindex] == NULL) {
      cerr << "NULL regions not printed\n";
    }
    else {
      reg_table->regions[stmt->rindex]->print();
    }
    */
    rap =  evaluate_expr(stmt->expr,reg_table,aut);
    rap->print();
    rap->free();
    global_free(rap);
    break;

  case PRINTSIZE_STMT: // print the region_array
    // a named region  only
    if (reg_table->regions[stmt->rindex] == NULL) {
      cerr << "Size of NULL regions not printed\n";
    }
    else {
      reg_table->regions[stmt->rindex]->print_size();
    }
    break;


  case PRINT_SIMPLIFIED_STMT: // print the region_array
    // beware: currently corrupts the automaton's location names.
    // need to restore them later.
//    cout << "Simplified output: \n" << flush;
    rap = evaluate_expr(stmt->expr, reg_table, aut);
    // this outputs the simplified list, using global control variables, and 
    // returns the simplified list.
    elim_all_locns = stmt->elim->elim_all;
    elim_locs_by_list = ! elim_all_locns;
    elim_proc_list = stmt->elim->name_list;
//    print_vars(elim_proc_list);
//    cout << "\n" << flush;
//    cout << "elim_all_locns: " << elim_all_locns << "\n" << flush;
    // also sets the elim var info, but this is irrelevant and does no harm
    aut->set_elim_info();
    aut->copy_location_names();
    rap->elim_locs_and_print();
    rap->free();
    aut->restore_location_names();
    break;

  case PRINT_STR_STMT: // print the region_array
      cout << stmt->rname << "\n" << flush;
    break;
  case TRACE_STMT: // generate trace to the expression
    rap1 = reg_table->regions[stmt->rindex];
    rap2 =  evaluate_expr(stmt->expr,reg_table,aut);
    aut->set_pre();
    rap1->generate_trace_to(rap2);
    rap2->free();
    global_free(rap2);
    break;
  case IF_THEN_STMT: case IF_THEN_ELSE_STMT:
//    cerr << "Doing if-then or if-then-else statement\n" << flush;
//     cout << "Evaluating Expression\n";
    if (result_of_Boolexpression(stmt->Boolexpr,reg_table,aut)) {
      execute_statements(stmt->thenstmt,reg_table,aut);
    }
    else {
      if (stmt->type == IF_THEN_ELSE_STMT) {
	execute_statements(stmt->elsestmt,reg_table,aut);
      }
    }
    break;
  case FREE_STMT: // free the region array
    if (reg_table->regions[stmt->rindex] == NULL) {
      cerr << "NULL regions not freed\n";
    }
    else {
      reg_table->regions[stmt->rindex]->free();
      reg_table->regions[stmt->rindex] = NULL;
    }
//    cerr << "Done with free of " << stmt->rindex << "\n" << flush;
    break;
  case ASSIGN_STMT: // figure out the region array for the expression
               // assign it to the variable
    rap = evaluate_expr(stmt->expr,reg_table,aut);
    if (reg_table->regions[stmt->rindex] != NULL) {
      reg_table->regions[stmt->rindex]->free();
      global_free(reg_table->regions[stmt->rindex]);
    }
    reg_table->regions[stmt->rindex] = rap;
//    cerr << "Done with assign to " << stmt->rindex << "\n" << flush;
    break;
  case WHILE_STMT:
    while (result_of_Boolexpression(stmt->Boolexpr,reg_table,aut)) {
      execute_statements(stmt->loopstmt, reg_table, aut);
      cerr << "." << flush;
      cout << "." << flush;
    }
    cerr << "\n" << flush;
    cout << "\n" << flush;
    break;
  default:
    cerr << "Illegal call of exec_spec(), with type " << stmt->type << "\n";
    cerr << flush;
    break;
  }
}




void execute_statements(struct statement *stmt, 
			struct region_table *reg_table, automaton *aut) {
  if (stmt != NULL) {
    execute_statement(stmt,reg_table,aut);
    execute_statements(stmt->next,reg_table,aut);
  }
}


