#ifndef __POLY_H
#define __POLY_H


/*
    - for a system of constraints :

      "1 a1 a2 ... an b" means "a1.x1 + a2.x2 + ... + an.xn +b >= 0"
      "0 a1 a2 ... an b" means "a1.x1 + a2.x2 + ... + an.xn +b  = 0"

    - for a system of generators :

      "1 a1 a2 ... an 0" is the ray [a1 a2 ... an]
      "1 a1 a2 ... an b" is the vertex [a1/b a2/b ... an/b]
      "0 a1 a2 ... an 0" is the line [a1 a2 ... an]

  Any matrix ends with the row "1 0 0 ... 0 1".
*/


#include<stdlib.h>
// #include<stream.h>

// g++ complains for not including stdio.h and iostream.h
#include<stdio.h>
#include<iostream.h>
#include<iomanip.h>
#include<math.h>

#define true 1
#define false 0
#define nil 0

typedef char Bool;

// set up the coefficient type 
#ifdef __COEFF_IS_INT
  #define COEFF_TYPE int
#endif
#ifdef __COEFF_IS_LONG_LONG_INT
  #undef __COEFF_IS_INT
  #define COEFF_TYPE long long int
#endif

#if !defined(COEFF_TYPE)
  #define  COEFF_TYPE int
#endif

typedef COEFF_TYPE coeff;


extern int overflow_avoid_mode;

class vector {
public:
  int size;
  coeff *p;


  vector();
  vector(int s);
  vector(const vector& vec);
  void free();
  void print();

// added HWT, deleted 5/5/95
//  void operator=(const vector& vec);
// added via PKPK, 9/19/95
   coeff& operator[](int i) {return p[i];}

};


void Gcd(coeff a, coeff b, coeff *g);
void Vector_Gcd(coeff *vec, int size, coeff &g);
void Vector_Normalize(coeff* vec, int size, int dimension);


class matrix {
  coeff *p_init;
public:
  int nbrows;
  int nbcolumns;
  coeff **p;
  Bool nonstrict;  // added HWT, 9/26/95, only used in intersections
  matrix();
  matrix(const matrix& m);            // copy constructor
  matrix(int nr, int nc);
  matrix(int nr, int nc, coeff* tab);
  matrix(int nr);
  void free();
  void print(ostream& outf=cout);
  void print(int, ostream& outf=cout);
  void ptprint(int form, ostream& outf=cout);

  void ptprintrow(int r, char **vnames, Bool *nontrivp, int form, ostream& outf=cout);
  void pretty_print(char **names, int form, ostream& outf=cout);
  matrix* share();
  matrix* copy();
  void operator=(const matrix& m);

  // added HWT
  // set to matrix with all rows = 0, except for final row 1 0000 1
  void blank();
  void set_last_row();
  Bool eps_present(int eps_pos);
  Bool is_strict();  // whether has any strict inequalities, i.e. 2 in first col
  Bool nplus1_is_strict(int k); // assumes in nplus1 form, whether has strict ineq., param is pos of eps_slack var
  Bool nplus1_def_nonstrict(int k) ; // yields whether n+1 form definitely represents a nonstrict matrix, possibly empty n+1 form.

  // return Gcd of entire matrix of frame's pts
  coeff gcd_of_frame_matrix();  
  coeff gcd_of_constr_matrix();

  // divide is boolean, whether to divide by scalar or multiply
  void scale_frame_matrix(coeff scalar, int divide);
  void scale_constr_matrix(coeff scalar, int divide);
  void print_simplified();  // simplify matrix, then print it
};


extern matrix* Read_Matrix();
extern vector* Read_Vector();

void my_new_handler();

// extern void transform(matrix mat, matrix *result, Bool constraints_wanted);
/* translation constraints <--> frame, according to 'constraint_wanted' */
extern void transform(matrix& m, matrix *n, matrix *result, Bool constraints_wanted, int known);
/* the same when a part of the result is already known */

class lib_polyhedron{
  int dimension;
  matrix C;
  matrix F;
  Bool frame_available;
  Bool constraints_available;
public:
  lib_polyhedron(matrix& mat, Bool constraint); // the constructor, from any representation
  lib_polyhedron(lib_polyhedron& p);           // copy constructor
  // ~lib_polyhedron();
  matrix *constraints(); // returns (after possible computation) the constraints
  matrix *frame();       // returns (after possible computation) the frame
  Bool empty;
  void print();
  void frame_print();
  void ptprint(ostream& outf=cout);
  lib_polyhedron *var_increment(int var_index); // increment a variable, in both representations
  lib_polyhedron *var_decrement(int var_index); // decrement a variable, in both representations
  lib_polyhedron *var_reset(int var_index);     // reset a variable (only in the frame)
  lib_polyhedron *var_assign(int var_index, vector *exp); 
                                           // assign a variable with a linear expression
  lib_polyhedron *add_elt(vector *elt);        // add a generating element
  lib_polyhedron *intersection(matrix c);  // intersects with new constraints
                                       // n.b. kills old poly

  lib_polyhedron *hull(matrix f);         // convex union with other frame
  void free();
  lib_polyhedron *copy();
  lib_polyhedron *minimize(); // minimization of both representations
  lib_polyhedron *add_a_null_variable(); // one more variable, initialized to 0
  lib_polyhedron *delete_variable(int d); // suppress one variable
  friend lib_polyhedron* empty_lib_polyhedron(int nbcolumns);


  void operator=(lib_polyhedron& p);

  /* added by HWT */
  // set frame invalid, and free memory if frame was being used.
  void invalidate_frame();
  void invalidate_constraints();



              // existentially quantify var vnr
              // out of a constraints with nr_extra_vars (for primed ass. vars)
	      // destroys old poly
  lib_polyhedron *existent(int vnr, int nr_extra_vars);     
              // existentially quantify nr_exist_vars vars, found in indices
	      // destroys old poly
  lib_polyhedron *existent(int nr_exist_vars, int *indices);
	      // destroys old poly
  lib_polyhedron *CopyColumn( int i2, int i3);
	      // destroys old poly
  lib_polyhedron *copy_extra_to_old(int nr_keep_cols, int nr_copy_cols, 
	int *index_p);
	      // returns new poly, original kept intact
  lib_polyhedron *vaporize_variable(int v); // remove column for this variable
                                        // no regard for semantic of var.
                                        // v is var_nr, not column number
	      // returns new poly, original kept intact
  lib_polyhedron *vaporize_variables(int nr_v); // remove column for last
                                        // nr_v vars.
                                        // no regard for semantic of var.
  lib_polyhedron *hull(lib_polyhedron *p); // takes hull, leaving p intact
                                   // but object destroyed

                                   // return fresh copy of poly. with
                                   // added constraint from row rownr of matrix
                                   // M.  Use negated constraint iff 
                                   // negated = true
  lib_polyhedron *add_row(matrix *mp, int rownr, Bool negated);
  int dim();  // returns dimension

  /* 12/96 */
  Bool contained_in(lib_polyhedron * p2);  // same as operator <=

  Bool contains_origin(); // yields whether poly contains the origin

};

Bool operator == (lib_polyhedron& p1, lib_polyhedron& p2);     // comparison
Bool operator <= (lib_polyhedron& p1, lib_polyhedron& p2);     // inclusion
lib_polyhedron *widenning (lib_polyhedron p1, lib_polyhedron p2); // standard widenning
lib_polyhedron *limited_widenning (lib_polyhedron p1, lib_polyhedron p2, matrix cont);
                                // widenning upto given constraints
lib_polyhedron *bounded_widenning (lib_polyhedron p1, lib_polyhedron p2, matrix cont);
                                // widenning upto given constraints and intervals

Bool check_empty(matrix* F);  // put here so publicly available



inline coeff sign_of(coeff m) {
  if (m<0) return -1;
  else if (m==0) return 0;
  else return 1;
}


inline coeff 
mymult(coeff m1, coeff m2) {
  coeff res = m1 * m2;
  int corr_res_sign;
  int sign_m1 = sign_of(m1);

  if (sign_m1 == 0) {
    corr_res_sign = 0;
  }
  else {
    int sign_m2 = sign_of(m2);
    if (sign_m2 == 0) {
      corr_res_sign = 0;
    }
    else {
      if (sign_m1 == sign_m2 ) {
	corr_res_sign = 1;
      }
      else {
	corr_res_sign = -1;
      }
    }
  }
      
  if (corr_res_sign != sign_of(res)) {
    cerr << "Library overflow error in multiplication\n";

cerr << "trying to multiply " << m1 << " and " << m2 << endl;


    if (overflow_avoid_mode < 2) {
      cerr << "Try overflow avoidance modes, -o1, -o2\n";
    }
    exit (1);
    return(1); // to keep CC compiler happy
  }
  else {
    return res;
  }
}


inline coeff 
myadd(coeff a1, coeff a2) {
  coeff res;
  int sign_of_a2 = sign_of(a2);

  res = a1+a2;
  if (sign_of(a1) == sign_of_a2) {
    if (a2!=0) {
      // cannot both be 0,  so either both pos or both neg
      if (sign_of(res) != sign_of_a2) {
	cerr << "Library overflow error in addition (incorrect sign of result)\n";
//	cerr << "Computes " << *sum << " + " << inc << " = " << res << "\n";
	exit (1);
      }
      else if (sign_of_a2 > 0) {
	if (res <= a1 || res <= a2) {
	  cerr << "Library overflow error in addition\n";
	  exit(1);
	}
      }
      else {
	// (sign_of_a2 < 0)
	if (res >= a1 || res >= a2) {
	 cerr << "Library overflow error in addition\n";
	  exit(1);
	}
      } 
    }
  }
  return(res);
}
	


#endif
