// header file for regions,  symbolic sets of states in a hybrid automaton
//
// used by pre-post and the verifier

#ifndef __REGION_H
#define __REGION_H

#include "poly.H"          
#include "extrapoly.H"    
#include "automaton.H"
#include "predicate.h"

class region {
 public:
  location * locnp;
  polyhedron * polyp;
  
  // for reachability and generation of error traces
  int reg_id;            // this region's id
  int pred_reg_id;       // id of predecessor region
  transition *trans_in;  // transition used to get from pred_reg_id to reg_id
  

  region * next;

  region();
  region(region & r);                  // copy data, but set .next to NULL
  // doesn't copy data for poly, just takes the ptr.
  region(location *locp, polyhedron *pp);

  // return fresh regions
  region *intersect(region * r);
  region *intersect(polyhedron * p);
  // intersect object's list with argument's list
  region *list_intersect_list(region *rp);  
  // intersect object with argument's list
  region* intersect_list(region* rp);
  Bool is_empty();               // is a single region empty
  Bool is_empty_given_eps_pos(int k);
  Bool is_empty_list();          // is the list pted to all empty poly's
  Bool is_contained_in(region& r);      
  Bool is_equal(region& r);             

  // returns exactly whether object continaed in rp's list
  Bool exact_is_in(region * rp); 
  // returns exactly whether states in object's list of regions are equivalent
  // to states in rp's list of regions
  Bool exact_is_equal(region *rp);

  region * copy();             /* copy the pointers only */
  region *fresh_copy();        /* make ptrs to NEW poly  */
  region *fresh_copy(int idnr);/* make ptrs to NEW poly  */
  region *fresh_copy_list();   /* makes  copy (data too) of entire list */
  region * insert(region * r);


  void print();                   // print an entire list
  void minimize();                // minimize poly
  void min_print();               // minimize poly, then print
  void list_print();              // print list, each locn name only once
  void single_print_0(Bool print_name);  // print single item at head of list
  void print_0(Bool print_name, Bool print_list);  // print single item at head of list
  void single_print();            // print single item at head of list
  void single_full_print();       // includes trans in, reg id, etc
  int list_size();
  void operator = (region& r); 
  region * cts_evolve_pre();
  region * cts_evolve_post();
                                   // old version do cts, then events.
  region * old_pre();              /* return list of pre regions, for trans */
  region * old_post();             /* return list of post region, for trans */
                                   // usual version does event then cts
                                   // to use these, assume you've done cts
                                   // of calling region
  region * pre();                  /* return list of pre regions, for trans */
  region * post();                 /* return list of post region, for trans */
                                   /* full version includes cts evolve of object */
  region * full_pre();             /* return list of pre regions, for trans */
  region * full_post();            /* return list of post region, for trans */

/*  region *hull(polyhedron *pp); */   /* return ptr to region for hull */

  void restrict_to_invariant();

  region * simplify();             // simplify a list of regions via region_or


  /* for closure diff, and exact containment and exact equality over unions */
  /* assume always that the regions contain the same locn info */
                                   // add constraint from row rownr of rp's
                                   // constraint matrix to object's constraints
  region * add_row(region *rp, int rownr); 
                                   // closure diff of single region's 
                                   //  rownr'th constraint


  Bool is_in_list(region *rp);         // is single object in list?
  Bool is_list_in_list(region *rp);    // is list of objects in list rp?
  Bool is_list_equal_list(region *rp); // is list = list?

                                   // free single elt only.
  void free();                     // deallocate space and matrix too.
  void free_list();                // free all elts on list.

                                   // return ptr to new. old is maintained
  region * cts_evolve(enum directions dirn);
  void elim_loc_name_for_proc(int proc_nr);

                                  // assign new IDs to entire list of regions
  void assign_reg_ids();          // use global max_reg_id

  region * single_event_succ(int i, enum directions dirn);
  region * event_succ(enum directions dirn);

  void remove_empties();          // from list 

  Bool is_non_strict();           // whether the poly is nonstrict
  // make name for process proc_nr be blank.

  // minimize rows for all regions in the list with this head
  void row_minimize();

  // for difference
  region* complement();       // complement of a single element (?)
  region* complement_list();  // complement of a whole list
  region* minus(region* rp);  // object - argument  (entire lists)
  region* region_and(region* rp);  // and of object's list and argument's list
};


// this class used for 
class region_array {
public:
  int size;   // nr of locations (?).
              // size is the number of  ptrs to lists of regions
  region **regions;

  // added 2/96 for use as init region from which a reachable region
  // is derived
  region_array * init_regn;

  // added 9/96
  automaton *autp;

  // ----------------------------------------------------

  region_array();
  region_array(int sz);

  region_array(region_array *rega);  // copy constructor
  region_array(struct convex_predicate *cp, automaton *aut);
  region_array(int sz, region *reg);

  // added 12/96
  region_array(int sz, automaton* autp);

  // checks that the region arrays have the same sizes
  // exits if not.
  void check_sizes_match(region_array regarr, char *routine_name);

  // Boolean operator, based on weakdiff
  Bool ra_in(region_array *rap);
  Bool ra_equiv(region_array *rap);

  Bool ra_empty();

  // operates on object, old object is lost
  region_array * restrict_to_invariant();
  region_array * ra_and(region_array regarr);
  region_array * ra_or(region_array regarr);
  
  // compute the "or", and put new regions in ranew.
  // source object is lost.
  region_array * ra_or_new(region_array regarr, region_array **ranew);
  region_array * ra_pre();
  region_array * ra_post();
  /* full versions include cts evolve of object */
  region_array * ra_full_pre();
  region_array * ra_full_post();
  region_array * ra_succ (enum directions dirn, int fullsucc);
  region_array * ra_reach(enum directions  dirn);

  region_array * ra_hull();

  // given a region array, prepare it for reachability in direction dirn
  // i.e. restrict all regions to the invariant, take cts evolve in dirn?
  // destroys old object, returns ptr to itself (the new object).
  region_array * prepare_for_reach(enum directions dirn, int add_cts_evolve);

   // generate an error trace, given
   // 1) object regions array for the reached region, with reg_id's, etc
   // 2) final_ra, the regions array for the final regions
   // N.B. in object ra, pred_reg_id == -1 means initial region
   void generate_trace_to(region_array *final_ra);
   // both next procedures should be protected
   void print_trace(region **id_to_reg, region_array *int_ra,
		    Bool print_forward);
   void print_trace_from(region **id_to_reg,Bool print_forward,
			 int tar_reg_id,  region *pt, float *cum_time_p);

  // all these elimin procedure can destroy the original, return ptr to this
  region_array * eliminate_vars(int nr_elim_vars, int *indices);
  region_array * eliminate_vars(struct var_name *var_names);
  region_array * eliminate_non_params();
  region_array * eliminate_all_vars();
  region_array * eliminate_location_for_proc(int proc_nr);
  // temporary routine
  // for now, just eliminate all locations.
  // return region list for result
  // can destroy the object in process
  region * eliminate_locations();
  // return simplified list of regions with str in their location's name
  region * eliminate_locations_with_str(char *str);
  // return list for with all locations and non-params eliminated.
  region * eliminate_all();
  
  // returns region, with names retained, but simplified
  region * merge_locations_by_name();

  // depends on parameters, what elimination is done.
  // does both vars and locs, prints output after each step
  // may also alter location names in the automaton!
  region * output_elim();

  // print after eliminates the locations, according to globals
  //  elim_all_locs, elim_locs_by_list, elim_proc_list
  // leave object intact
  void elim_locs_and_print();
  // as above, but
  // 1) assumes eliminating all locations, and
  // 2) stores a copy of the expression in every location
  void elim_locs_and_store(automaton *aut);


  // return a list of regions for the simplification by eliminating all locns
  //   and non parameters
  // simplify each list of regions
  void simplify();

  // set all names to some constant (NULL)
  void make_names_constant();

  // add a list of regions to the region array
  // destroy the list in the process
  void add_to(region *rp);


  // remove empty items  from lists
  void remove_empties();

  // for exact containment, etc. cf ra_in, ra_equiv which use weakdiff
  //   and therefore only approximate.

  // following routines only work for fully nonstrict regions
                                                    // closure of diff of lists
  region_array * closure_diff(region_array *rap);   // per locn 
  Bool ra_exact_is_in(region_array *rap);           // exact containment check
  Bool ra_exact_is_equal(region_array *rap);        // exact equivalence check
  Bool ra_exact_is_lt(region_array *rap);           // exact < check
  Bool ra_exact_is_gt(region_array *rap);           // exact > check

  Bool ra_is_non_strict();  // checks whether every region is nonstrict

  void free();
  Bool is_empty();
  void print();
  void print_size();
  void ra_assign_reg_ids();

  // for difference
  region_array* ra_complement();
  region_array* ra_minus(region_array* rap);
};

Bool operator <= (region r1, region r2); // inclusion
region_array * init_region_array(automaton *autp, enum directions dirn);
region_array * final_region_array(automaton *autp, enum directions dirn);


region* true_region(location *locp, int sz);

// compute OR of the list of regions.
// add regions in regp2 to those in *reglist
// the original *reglist may be lost in computing the new list, but keep 
//   all data from regp2
void region_or(region **reglist, region *regp2);


// construct a region array corresponding to the disjunctive predicate
// of convex parts.
// autp is the reference automaton.  it's a product automaton.
region_array *
ra_for_pred(automaton *autp, struct convex_predicate *cp);


// temporary
void test_reached_set(region_array rega);



void automaton_simple_deadlock_check(automaton a);

#endif
